/*
	libcastor, see LICENSE for terms of use
*/

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>

#include "castor.h"

/* register access */
int memfd;
int isinit;
volatile uint32_t *memregs32;
volatile uint16_t *memregs16;
volatile uint8_t  *memregs8;
volatile uint16_t *uppermem;
uint32_t bkregs32[15];	/* backing up values */

/* library variables */
char error_buf[64];
int video_mode[2];
int pixel_width[2];
int layer_width[2];
int screen_h, screen_w;
volatile uint16_t *lc_fb0, *lc_fb1;		/* library repoints these to whatever the next frame will be */
volatile uint8_t *lc_8bpp_fb0, *lc_8bpp_fb1;
volatile uint32_t *lc_32bpp_fb0, *lc_32bpp_fb1;
volatile uint16_t *fb0_0, *fb0_1;	/* layer 0, buffer 0 : layer 0, buffer 1 (RGB) */
volatile uint16_t *fb1_0, *fb1_1;	/* layer 1, buffer 0 : layer 1, buffer 1 (RGB) */
volatile uint16_t *vb_0, *vb_1;		/* layer 0, buffer 0 : layer 0, buffer 1 (YUV video) */
uint16_t lc_palette[2][256];

/* internal prototypes */
static void lc_seterr(const char* err);

/* base lib functions */
int lc_init(void) {
	printf("Init libcastor... ");

	/* open /dev/mem to access registers */
	memfd = open("/dev/mem", O_RDWR);
	if(memfd < 0) {
		lc_seterr("Could not open /dev/mem!\n");
		goto ERROR_INIT_DEVMEM;
	}

	/* get access to the registers */
	memregs32 = (volatile uint32_t*)mmap(0, 0x20000, PROT_READ|PROT_WRITE, MAP_SHARED, memfd, 0xC0000000);
	if(memregs32 == (volatile uint32_t*)0xFFFFFFFF) {
		lc_seterr("Could not mmap hardware registers!\n");
		goto ERROR_INIT_MMAP;
	}
	memregs16 = (volatile uint16_t*)memregs32;
	memregs8 = (volatile uint8_t*)memregs32;

	/* map the upper memory */
	uppermem = (volatile uint16_t*)mmap(0, 0x12C000, PROT_READ|PROT_WRITE, MAP_SHARED, memfd, 0x3700000);
	if(uppermem == (volatile uint16_t*)0xFFFFFFFF) {
		lc_seterr("Could not mmap upper memory!\n");
		goto ERROR_INIT_MMAP_UM;
	}

	/* backup old register values to restore upon exit */
	bkregs32[0] = MLCADDRESS0;
	bkregs32[1] = MLCADDRESS1;
	bkregs32[2] = MLCCONTROL0;
	bkregs32[3] = MLCCONTROL1;
	bkregs32[4] = MLCLEFTRIGHT0;
	bkregs32[5] = MLCTOPBOTTOM0;
	bkregs32[6] = MLCLEFTRIGHT1;
	bkregs32[7] = MLCTOPBOTTOM1;
	bkregs32[8] = MLCBGCOLOR;
	bkregs32[9] = MLCHSTRIDE0;
	bkregs32[10] = MLCVSTRIDE0;
	bkregs32[11] = MLCHSTRIDE1;
	bkregs32[12] = MLCVSTRIDE1;
	bkregs32[13] = DPCCTRL1;
	bkregs32[14] = MLCSCREENSIZE;

	/* assign framebuffers */
	fb0_0 = (volatile uint16_t*)mmap(0, 0x4B000, PROT_READ|PROT_WRITE, MAP_SHARED, memfd, 0x3700000);
	fb0_1 = (volatile uint16_t*)mmap(0, 0x4B000, PROT_READ|PROT_WRITE, MAP_SHARED, memfd, 0x374B000);
	fb1_0 = (volatile uint16_t*)mmap(0, 0x4B000, PROT_READ|PROT_WRITE, MAP_SHARED, memfd, 0x3796000);
	fb1_1 = (volatile uint16_t*)mmap(0, 0x4B000, PROT_READ|PROT_WRITE, MAP_SHARED, memfd, 0x37E1000);
	lc_fb0 = fb0_1; /* assign initial framebuffers */
	lc_fb1 = fb1_1;
	lc_8bpp_fb0 = (uint8_t*) fb0_1;
	lc_8bpp_fb1 = (uint8_t*) fb1_1;
	lc_32bpp_fb0 = (uint32_t*) fb0_1;
	lc_32bpp_fb1 = (uint32_t*) fb1_1;
	memset((void*)fb0_0, 0x00, 0x4B000); /* clear framebuffers */
	memset((void*)fb0_1, 0x00, 0x4B000);
	memset((void*)fb1_0, 0x00, 0x4B000);
	memset((void*)fb1_1, 0x00, 0x4B000);

	lc_screensize(240, 320); /* set default screen orientation */
	lc_setbgcol(0x000000); /* set default background colour */
	lc_layerpos(0, 0, 0, 239, 319);	/* set default layer positions */
	lc_layerpos(1, 0, 0, 239, 319);
	lc_setlayer(0, false, false, false, false, RGB565); /* set default layer settings */
	lc_setlayer(1, true, false, false, false, RGB565);
	lc_flipfb(0);	/* set initial addresses in hardware */
	lc_flipfb(1);
	usleep(100000);

	/* generic library stuff */
	isinit = 1;
	atexit(lc_exit);

	printf("success!\n");
	return 0;

	ERROR_INIT_MMAP_UM:
	ERROR_INIT_MMAP:
		close(memfd);
	ERROR_INIT_DEVMEM:
	return -1;
}

void lc_exit(void) {
	/* if the library hasn't been init-ed, can't deinit */
	if(!isinit) {
		return;
	}
	printf("Exit libcastor... ");

	/* restore old register values */
	MLCADDRESS0 = bkregs32[0];
	MLCADDRESS1 = bkregs32[1];
	MLCCONTROL0 = bkregs32[2];
	MLCCONTROL1 = bkregs32[3];
	MLCLEFTRIGHT0 = bkregs32[4];
	MLCTOPBOTTOM0 = bkregs32[5];
	MLCLEFTRIGHT1 = bkregs32[6];
	MLCTOPBOTTOM1 = bkregs32[7];
	MLCBGCOLOR = bkregs32[8];
	MLCHSTRIDE0 = bkregs32[9];
	MLCVSTRIDE0 = bkregs32[10];
	MLCHSTRIDE1 = bkregs32[11];
	MLCVSTRIDE1 = bkregs32[12];
	DPCCTRL1 = bkregs32[13];
	MLCSCREENSIZE = bkregs32[14];

	lc_dirtylayer(0);
	lc_dirtylayer(1);
	lc_dirtymlc();

	memregs8 = NULL;
	memregs16 = NULL;
	memregs32 = NULL;
	close(memfd);
	isinit = 0;
	printf("success!\n");
}

int lc_isinit(void) {
	return isinit;
}

static void lc_seterr(const char* err) {
	if(strlen(err) <= 64) {
		printf("libpollux: %s\n", err);
		strcpy(error_buf, err);
	}
}

char* lc_geterr(void) {
	return error_buf;
}

/* video functions */
void lc_setfb(int layer, volatile uint16_t* set_to) {
	/* set absolute address for framebuffers */
	if(layer == 0) {
		lc_fb0 = set_to;
		lc_8bpp_fb0 = (uint8_t*) set_to;
		lc_32bpp_fb0 = (uint32_t*) set_to;
		if(set_to == fb0_0) {
			MLCADDRESS0 = 0x3700000;
		} else {
			MLCADDRESS0 = 0x374B000;
		}
	} else {
		lc_fb1 = set_to;
		lc_8bpp_fb1 = (uint8_t*) set_to;
		lc_32bpp_fb1 = (uint32_t*) set_to;
		if(set_to == fb1_0) {
			MLCADDRESS1 = 0x3796000;
		} else {
			MLCADDRESS1 = 0x37E1000;
		}
	}
	lc_dirtylayer(layer);
}

void lc_flipfb(int layer) {
	/* switch to the other buffer */
	static int current_0 = 0;
	static int current_1 = 0;

	if(layer == 0) {
		current_0 = !current_0;
		lc_setfb(0, current_0 ? fb0_1 : fb0_0);
	} else {
		current_1 = !current_1;
		lc_setfb(1, current_1 ? fb1_1 : fb1_0);
	}
}

void lc_setlayer(int layer, bool onoff, bool alpha, bool invert, bool trans, uint32_t mode) {
	/* set layer properties register */
	uint32_t temp;
	temp = 0;
	if(onoff)	temp |= BIT(5);
	if(alpha)	temp |= BIT(2);
	if(invert)	temp |= BIT(1);
	if(trans)	temp |= BIT(0);
	temp |= BIT(12);
	temp |= BIT(14);
	temp |= BIT(15);
	if(mode)	video_mode[layer] = mode;
	temp |= (video_mode[layer]<<16);

	if(layer == 0) {
		MLCCONTROL0 = temp;
	} else {
		MLCCONTROL1= temp;
	}
	lc_dirtylayer(layer);

	/* set stride based on pixel width*/
	switch(mode) {
		case RGB565:
		case BGR565:
		case XRGB1555:
		case XBGR1555:
		case XRGB4444:
		case XBGR4444:
		case XRGB8332:
		case XBGR8332:
		case ARGB1555:
		case ABGR1555:
		case ARGB4444:
		case ABGR4444:
		case ARGB8332:
		case ABGR8332:
			pixel_width[layer] = 2;
			break;
		case RGB888:
		case BGR888:
			pixel_width[layer] = 3;
			break;
		case ARGB8888:
		case ABGR8888:
			pixel_width[layer] = 4;
			break;
		case PTRGB565:
			pixel_width[layer] = 1;
			break;
		default:
			break;
	}
	lc_setstride(layer, pixel_width[layer], pixel_width[layer]*layer_width[layer]);
}

void lc_layerpos(int layer, int x1, int y1, int x2, int y2) {
	uint32_t temp_lr, temp_tb;
	temp_lr = (x1 << 16) | x2;
	temp_tb = (y1 << 16) | y2;

	if(layer == 0) {
		MLCLEFTRIGHT0 = temp_lr;
		MLCTOPBOTTOM0 = temp_tb;
	} else {
		MLCLEFTRIGHT1 = temp_lr;
		MLCTOPBOTTOM1 = temp_tb;
	}
	lc_dirtylayer(layer);

	layer_width[layer] = (x2-x1)+1;
}

void lc_setalpha(int layer, int value) {
	if(value < 0 || value > 15)
		return;

	if(layer == 0) {
		MLCTPCOLOR0 = (MLCTPCOLOR0&0xFFFFFF) | value << 28;
	} else {
		MLCTPCOLOR1 = (MLCTPCOLOR1&0xFFFFFF) | value << 28;
	}
	lc_dirtylayer(layer);
}

void lc_settranscol(int layer, uint32_t colour) {
	if(layer == 0) {
		MLCTPCOLOR0 = (MLCTPCOLOR0&0xFF000000) | (colour&0xFFFFFF);
	} else {
		MLCTPCOLOR1 = (MLCTPCOLOR1&0xFF000000) | (colour&0xFFFFFF);
	}
	lc_dirtylayer(layer);
}

void lc_setinvcol(int layer, uint32_t colour) {
	if(layer == 0) {
		MLCINVCOLOR0 = colour;
	} else {
		MLCINVCOLOR1 = colour;
	}
	lc_dirtylayer(layer);
}

void lc_dirtymlc(void) {
	MLCCONTROLT |= BIT(3);
}

void lc_dirtylayer(int layer) {
	if(layer == 0) {
		MLCCONTROL0 |= BIT(4);
	} else {
		MLCCONTROL1 |= BIT(4);
	}
}

#define FBIO_MAGIC			'D'
#define	FBIO_LCD_CHANGE_CONTROL		_IOW(FBIO_MAGIC, 90, unsigned int[2])
#define	LCD_DIRECTION_ON_CMD		5	/* 320x240 */
#define	LCD_DIRECTION_OFF_CMD		6	/* 240x320 */
void lc_screensize(int w, int h) {
	unsigned int send[2];
	int fb_fd = open("/dev/fb0", O_RDWR);
	send[1] = 0;
	/* alter MLC to rotate the display */
	if(w == 320 && h == 240) {
		screen_w = 320; screen_h = 240;
		send[0] = LCD_DIRECTION_ON_CMD;
	} else if(w == 240 && h == 320) {
		screen_w = 240; screen_h = 320;
		send[0] = LCD_DIRECTION_OFF_CMD;
	} else {
		lc_seterr("Tried to set invalid screen size\n");
	}
	/* send command to display controller */
	ioctl(fb_fd, FBIO_LCD_CHANGE_CONTROL, &send);
	close(fb_fd);
	/* apply the MLC changes */
	MLCSCREENSIZE = ((h-1)<<16) | (w-1);
	lc_dirtymlc();
}

void lc_setbgcol(uint32_t colour) {
	/* colour to be displayed where no layers cover */
	MLCBGCOLOR = colour;
	lc_dirtymlc();
}

void lc_setstride(int layer, int hs, int vs) {
	/* set how many bytes the MLC is supposed to read */
	if(layer == 0) {
		MLCHSTRIDE0 = hs;
		MLCVSTRIDE0 = vs;
	} else {
		MLCHSTRIDE1 = hs;
		MLCVSTRIDE1 = vs;
	}
	lc_dirtylayer(layer);
}

void lc_setdither(int r, int g, int b) {
	/* set ditheting - does this actually work? */
	DPCCTRL1 = (DPCCTRL1&0x3F) | (b<<4) | (g<<2) | r;
}

#define SYS_CLK_FREQ 27
void lc_setclock(int speed) {
	unsigned  long v;
	unsigned mdiv, pdiv=9, sdiv=0;

	mdiv= (speed * pdiv) / SYS_CLK_FREQ;
	mdiv &= 0x3FF;
	v= pdiv<<18 | mdiv<<8 | sdiv;

	PLLSETREG0 = v;
	PWRMODE |= 0x8000;
}

uint32_t lc_getbuttons(void) {
	return ~((GPIOCPAD << 16) | GPIOBPAD);
}


time_t lc_gettime() {
	time_t new;

	RTCCTRL |= BIT(0);
	new = (time_t)RTCCNTREAD;
	RTCCTRL &= ~BIT(0);

	stime(&new);

	return new;
}

void lc_settime(time_t setto) {
	if(RTCCTRL & BIT(4))
		lc_seterr("Could not set time.\n");
	else
		RTCCTRL |= BIT(0);
		RTCCNTWRITE = (int32_t)setto;
		RTCCTRL &= ~BIT(0);
}

void lc_setpalette(int layer) {
	int i;
	if(layer == 0) {
		for (i= 0 ; i < 256; i++) {
			MLCPALETTE0 = i<<24 | lc_palette[0][i];
		}
	} else {
		for (i= 0 ; i < 256; i++) {
			MLCPALETTE1 = i<<24 | lc_palette[1][i];
		}
	}

	lc_dirtylayer(layer);
}

