#ifndef __LIBCASTOR_H__
#define __LIBCASTOR_H__
#include <stdint.h>
#include <stdbool.h>
#include <time.h>
#include "polluxregs.h"

/*
	kernel memory map
		0x3B00000 - 59MB - Start of sound buffer
		0x3700000 - 55MB - Start framebuffer data
		0x3000000 - 48MB - Start MPEG decoder crap
	IGNORE THIS FOR NOW > libpollux memory map < IGNORE THIS FOR NOW
		0x3000000-0x304AFFF : First RGB buffer, layer 0 (300K, max needed for 32bit colour at 320x240)
		0x374B000-0x3095FFF : Second RGB buffer, layer 0 (300K, max needed for 32bit colour at 320x240)
		0x3796000-0x30E0FFF : First RGB buffer, layer 1 (300K, max needed for 32bit colour at 320x240)
		0x37E1000-0x312BFFF : Second RGB buffer, layer 1 (300K, max needed for 32bit colour at 320x240)
		0x312C000-?	    : YUV 4:2:0 buffer
*/
#ifndef NULL
	#define NULL	0
#endif
#ifndef __bool_true_false_are_defined
	typedef uint32_t bool;
	#define true	1
	#define false	0
#endif
extern volatile uint16_t *lc_fb0, *lc_fb1;
extern volatile uint8_t *lc_8bpp_fb0, *lc_8bpp_fb1;
extern volatile uint32_t *lc_32bpp_fb0, *lc_32bpp_fb1;
extern uint16_t lc_palette[2][256]; /* palettes */
#define RGB24TO16(R, G, B)	(((((R)&0xF8)<<8)|(((G)&0xFC)<<3)|(((B)&0xF8)>>3)))

/*
	lc_abstorel	Converts and absolte register address to one relative to 0xC0000000)
	lc_absupmem	Converts and absolte register address to one relative to 0x03000000)
*/
#define lc_abstorel(addr, bits)	((addr-0xC0000000)>>(bits>>4))
#define lc_absupmem(addr)	((addr-0x03000000)>>(16>>4))
/*
	lc_peekreg(address, bits)	Returns the value stored in register at 'address'

	notes:
		pass the absolute address to this function (e.g. 0xC0004000, NOT 0x4000)
	returns:
		value held in the register
*/
#define lc_peekreg(address, bits) ( \
	bits == 32 ? memregs32[lc_abstorel(address, 32)] : \
	(bits == 16 ? memregs16[lc_abstorel(address, 16)] : memregs8[lc_abstorel(address, 8)]) \
)
/*
	lc_setreg(address, bits)	Stores the value in register at 'address'

	notes:
		pass the absolute address to this function (e.g. 0xC0004000, NOT 0x4000)
	returns:
		nothing
*/
#define lc_setreg(address, bits, data) ( \
	bits == 32 ? memregs32[lc_abstorel(address, 32)]=data : \
	(bits == 16 ? memregs16[lc_abstorel(address, 16)]=data : memregs8[lc_abstorel(address, 8)]=data) \
)

/*
	lc_init()	Starts the library

	returns:
		0 on success
		-1 on failure (details accessible from lc_geterr)
*/
extern int lc_init(void);

/*
	lc_exit()	Quits the library

	returns:
		nothing
*/
extern void lc_exit(void);

/*
	lc_isinit()	Checks if libcastor has been init-ed

	returns:
		1 or 0
*/
extern int lc_isinit(void);

/*
	lc_geterr()	Gets the last error message issued by the library

	returns:
		error message
*/
extern char* lc_geterr(void);

/*
	lc_setfb(layer, set_to)	Sets the current framebuffer to set_to

	returns:
		nothing
*/
extern void lc_setfb(int layer, volatile uint16_t* set_to);

/*
	lc_flipfb(layer)	Flips to the other buffer for a particular layer

	returns:
		nothing
*/
extern void lc_flipfb(int layer);

/*
	lc_setlayer(layer, onoff, alpha, invert, trans, mode)	Sets layer properties

	returns:
		nothing
*/
extern void lc_setlayer(int layer, bool onoff, bool alpha, bool invert, bool trans, uint32_t mode);

/*
	lc_layerpos(layer, x1, y1, x2, y2)	Sets layer position

	returns:
		nothing
*/
extern void lc_layerpos(int layer, int x1, int y1, int x2, int y2);

/*
	lc_setalpha(layer, value)	Sets layer alpha value (15 = opaque, 0 = transparent)

	returns:
		nothing
*/
extern void lc_setalpha(int layer, int value);

/*
	lc_settranscol(layer, colour)	Sets layer transparency colour (RGB888)

	returns:
		nothing
*/
extern void lc_settranscol(int layer, uint32_t colour);


/*
	lc_setinvcol(layer, colour)	Sets layer inversion colour (RGB888)

	returns:
		nothing
*/
extern void lc_setinvcol(int layer, uint32_t colour);

/*
	lc_dirtymlc()	Sets the dirty flag for the MLC

	returns:
		nothing
*/
extern void lc_dirtymlc(void);

/*
	lc_dirtylayer(layer)	Sets the dirty flag for the layer

	returns:
		nothing
*/
extern void lc_dirtylayer(int layer);

/*
	lc_screensize(width, height)	Sets the screen size

	returns:
		nothing
*/
extern void lc_screensize(int w, int h);

/*
	lc_setbgcol(colour)	Sets the background colour

	returns:
		nothing
*/
extern void lc_setbgcol(uint32_t colour);

/*
	lc_setstride(layer, hs, vs)	Sets stride registers

	returns:
		nothing
*/
extern void lc_setstride(int layer, int hs, int vs);

/*
	lc_setdither(r, g, b)	Sets dithering

	returns:
		nothing
*/
extern void lc_setdither(int r, int g, int b);

/*
	lc_setclock(speed)	Sets ARM926TE-J clock speed

	returns:
		nothing
*/
extern void lc_setclock(int speed);

/*
	lc_getbuttons()

	returns:
		button presses (& with a constant to test)
*/
extern uint32_t lc_getbuttons(void);
enum {
	BTN_A = BIT(20),
	BTN_B = BIT(21),
	BTN_X = BIT(22),
	BTN_Y = BIT(23),
	BTN_L = BIT(7),
	BTN_R = BIT(6),
	BTN_MENU = BIT(9),
	BTN_SELECT = BIT(8),
	BTN_LEFT = BIT(16),
	BTN_RIGHT = BIT(17),
	BTN_UP = BIT(18),
	BTN_DOWN = BIT(19),
	BTN_VOLUP = BIT(10),
	BTN_VOLDOWN = BIT(11)
};

/*
	lc_settime(time_t)	Set's the RTC time
	lc_gettime()		Returns the current time
*/
extern void lc_settime(time_t setto);
extern time_t lc_gettime();

/*
	lc_setpalette()	Uploads palette data to hardware

	returns:
		nothing
*/
extern void lc_setpalette(int layer);

#endif
